/*                               -*- Mode: C -*- 
 * 
 * ihsh.c -- 
 * 
 * Copyright (c) 2003 Arne Helme. All rights reserved.
 * Copyright (c) 2003 Juergen Sander. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Arne Helme.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: ihsh.c,v 1.7 2003/10/22 14:06:50 arne Exp $
 * 
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <expat.h>

#include <ih/tree.h>
#include <ih/tree_xml.h>

#include "ihsh.h"
#ifdef REMEDY
#include "ihsh_ars.h"
#endif
#include "ihsh_cfg.h"
#include "ihsh_cmds.h"
#include "ihsh_misc.h"
#include "ihsh_version.h"
#include "ihsh_xml.h"

#ifndef IHSH_COPYRIGHT
#define IHSH_COPYRIGHT "Copyright (c) 2003 Arne Helme. All rights reserved."
#endif


/* **************************************************************
 * Global variables
 * ************************************************************** */

/* Debug variable */
int debug = FALSE;

/* Program name */
char *progname;


/* **************************************************************
 * Functions
 * ************************************************************** */

/* Function to show program usage */
static void usage(void)
{
    fprintf(stderr, 
	    "usage: %s [-d][-c cfgfile][-f file][-o fname][-x fname]\n", 
	    progname);
    exit(EXIT_FAILURE);
}


/* **************************************************************
 * Main program
 * ************************************************************** */

/* Main program */
int main(int argc, char *argv[])
{
    ualloc_t *ua;
    ih_tree_t *tree;
    int		ch;
    char	*optstr;
    int		fflag, oflag, xflag;
#ifdef REMEDY
    int		rflag, sflag;
    char	*arsentryid;
#endif
    FILE	*cmdinfp;
    FILE	*xmlinfp;
    FILE	*xmloutfp;
    char	*cfgfname;

    /* Set program name */
    progname = argv[0];

    /* Print banner */
    fprintf(stderr, "%s %s. %s\n", progname, IHSH_VERSION, IHSH_COPYRIGHT);

    /* Parse command line arguments */
#ifdef REMEDY
    optstr = "c:df:o:r:sx:"; /* r ars remedy */
    arsentryid = NULL;
    rflag = sflag = FALSE;
#else
    optstr = "c:df:o:x:";
#endif
    cmdinfp = stdin;
    xmloutfp = stdout;
    fflag = oflag = xflag = FALSE;
    cfgfname = NULL;
    while ((ch = getopt(argc, argv, optstr)) != -1) {
        switch (ch) {
	case 'c': /* Alternative configuration file */
	    cfgfname = optarg;
	    break;
	case 'd': /* Debug mode */
	    debug = TRUE;
	    break;
	case 'f': /* Batch mode, read from file */
	    fflag = TRUE;
	    cmdinfp = fopen(optarg, "r");
	    if (cmdinfp == NULL) {
		perror(progname);
		fprintf(stderr, "%s: could not open file: %s for reading\n",
			progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'o': /* Write XML output to file */
	    oflag = TRUE;
	    xmloutfp = fopen(optarg, "w");
	    if (xmloutfp == NULL) {
		perror(progname);
		fprintf(stderr, "%s: could not open file: %s\n for writing",
			progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
#ifdef REMEDY
	case 'r': /* ARS Remedy retrieve (get) mode */
	    rflag = TRUE;
	    arsentryid = optarg;
	    break;
	case 's': /* ARS Remedy set mode */
	    sflag = TRUE;
	    break;
#endif
	case 'x': /* Read IODEF XML document from file */
	    xflag = TRUE;
	    xmlinfp = fopen(optarg, "r");
	    if (xmlinfp == NULL) {
		perror(progname);
		fprintf(stderr, "%s: could not open file: %s for reading\n",
			progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
        case '?':
        default:
            usage();
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (xflag == TRUE && fflag == TRUE) {
	usage();
    }

    /* Read IHSH configuration file */
    if (cfgfname == NULL) {
	char *home;
	
	home = getenv("HOME");
	if (home == NULL) {
	    fprintf(stderr, "%s: warning, the HOME environment not found!\n", 
		    progname);
	    home = "";
	}
	cfgfname = (char *) malloc(strlen(home) + strlen("/") + 
				   strlen(".ihshrc") + 1);
	if (cfgfname == NULL) {
	    fprintf(stderr, "%s: could not allocate memory for cfgfname\n",
		    progname);
	    exit(EXIT_FAILURE);
	}
	(void) sprintf(cfgfname, "%s/%s", home, ".ihshrc");
    }
    errno = 0;
    if (ihsh_cfg_readfile(cfgfname) < 0) {
	if (errno == 0) {
	    /* Syntax error in config file: kill the program */
	    exit(EXIT_FAILURE);
	}
    }

#ifdef REMEDY
    /* Initialize ARS/Remedy Module */
    if (rflag == TRUE || sflag == TRUE) {
	if (ihsh_ars_init() < 0) {
	    fprintf(stderr, "%s: could not initialize ARS/Remedy module\n",
		    progname);
	    exit(EXIT_FAILURE);
	}
    }
#endif

    /* Initialize LibIH memory allocation pool */
    ua = ualloc_create(NULL, NULL, 0, NULL, NULL, NULL);
    if (ua == NULL) {
	fprintf(stderr, "could not initialize LibIH memory pool\n");
	exit(EXIT_FAILURE);
    }

    /* Create initial LibIH IODEF XML tree */
    tree = ih_tree_create(ua, IH_TREE_T_IODEF);
    if (tree == NULL) {
	fprintf(stderr, "could not create initial XML tree\n");
	exit(EXIT_FAILURE);
    }

    /* Read in XML file */
    if (xflag == TRUE) {
	if (ihsh_xml_fread(&tree, xmlinfp, ua) < 0) {
	    fprintf(stderr, "could not read XML file\n");
	    exit(EXIT_FAILURE);
	}
	fclose(xmlinfp);
    }
    
#ifdef REMEDY
    /* If invoked with an ARS entry ID, invoke ARS getentry command */
    if (rflag == TRUE) {
	if (ihsh_ars_getentry(tree, arsentryid) < 0) {
	    fprintf(stderr, "Could not get ARS entry\n");
	    exit(EXIT_FAILURE);
	}
    }
#endif

    /* Program main loop */
    if (xflag == FALSE
#ifdef REMEDY
	&& rflag == FALSE
#endif
	) {
	ihsh_cmd_fproc(cmdinfp, &tree, ua);
    }

#ifdef REMEDY
    /* Invoke ARS set command */
    if (sflag == TRUE) {
	if (ihsh_ars_setentry(tree) < 0) {
	    fprintf(stderr, "Could not set ARS entry\n");
	    exit(EXIT_FAILURE);
	}
    } else
#endif

	if (ihsh_xml_fwrite(tree, xmloutfp) < 0) {
	    fprintf(stderr,  "Error: could not write XML data\n");
	    exit(EXIT_FAILURE);
	} else {
	    if (xmloutfp != stdout) {
		fclose(xmloutfp);
	    }
	}

    /* Clean up and exit */
    exit(EXIT_SUCCESS);
    /* NOTREACHED */
}

