/*                               -*- Mode: C -*- 
 * 
 * ihsh_ars.c -- ARS Remedy module for IHSH
 * 
 * Copyright (c) 2003 Arne Helme. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Arne Helme.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: ihsh_ars.c,v 1.5 2003/10/22 13:48:46 arne Exp $
 * 
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>

#include <ar.h>
#include <arfree.h>
#include <arextern.h>

#include <ih/tree.h>

#include "ihsh.h"
#include "ihsh_ars.h"
#include "ihsh_ars_iodef.h"
#include "ihsh_ars_misc.h"
#include "ihsh_cfg.h"

/* **************************************************************
 * IHSH ARS Remedy module local variables
 * ************************************************************** */

/* Hostname of the ARS server */
static char *ars_server = NULL;

/* Port the ARS server is listening to */
static int ars_port = 0;

/* ARS user */
static char *ars_user = NULL;

/* ARS password */
static char *ars_passwd = NULL;

/* Boolean variable to hold the IHSH ARS module initialization status */
static int ars_initialized = FALSE;

/* **************************************************************
 * IHSH ARS Remedy local functions
 * ************************************************************** */

/* **************************************************************
 * IHSH ARS Remedy functions
 * ************************************************************** */

/* Function to initialize the IHSH ARS/Remedy module */
int ihsh_ars_init(void)
{
    char *cfgstr;

    /* Inform user of ARS/Remedy availability */
    fprintf(stderr, "%s ARS v5.1 driver enabled and initialized\n", progname);

    /* Fetch ars_debug entry from IHSH configuration file */
    cfgstr = ihsh_cfg_match_str("ars_debug");
    if (cfgstr != NULL) {
	if (!strcmp(cfgstr, "TRUE") ||
	    !strcmp(cfgstr, "true") ||
	    !strcmp(cfgstr, "YES") ||
	    !strcmp(cfgstr, "yes")) {
	    fprintf(stderr, "%s: ARS: debug mode enabled\n", progname);
	    ars_debug = TRUE;
	}
    }

    /* Fetch ars_server entry from configuration file */
    cfgstr = ihsh_cfg_match_str("ars_server");
    if (cfgstr == NULL) {
	ARS_DEBUG("'ars_server' not set in IHSH config file");
	return -1;
    }
    ars_server = cfgstr;
    ARS_DEBUG("server: %s", ars_server);

    /* Fetch ars_port entry from IHSH configuration file */
    errno = 0;
    if (ihsh_cfg_match_int("ars_port", &ars_port) < 0) {
	ARS_DEBUG("'ars_port' not set in IHSH config file (or ERANGE)");
	return -1;
    }
    if (ars_port < 0 || ars_port > 65535) {
	ARS_DEBUG("'ars_port' must be 0 <= x < 65535");
	return -1;
    }
    ARS_DEBUG("port: %d", ars_port);

    /* Fetch ars_user entry from configuration file */
    cfgstr = ihsh_cfg_match_str("ars_user");
    if (cfgstr == NULL) {
	ARS_DEBUG("'ars_user' not set in IHSH config file");
	return -1;
    }
    ars_user = cfgstr;
    ARS_DEBUG("user: %s", ars_user);

    /* Fetch ars_passwd entry from configuration file */
    cfgstr = ihsh_cfg_match_str("ars_passwd");
    if (cfgstr == NULL) {
	ARS_DEBUG("'ars_passwd' not set in IHSH config file");
	return -1;
    }
    ars_passwd = cfgstr;
    ARS_DEBUG("password: xxxxxx");

    /* Flag module as initialized and return successfully */
    ars_initialized = TRUE;
    return 0;
}

/* Function to get entry from ARS/Remedy server */
int ihsh_ars_getentry(ih_tree_t *tree, char *entryid)
{
    ARControlStruct  control;
    ARStatusList     status;
    int ret;

    /* Check if IHSH ARS module is initialized */
    if (ars_initialized == FALSE) {
	ARS_DEBUG("ihsh_ars_getentry: module not initialized!");
	return -1;
    }

    ARS_DEBUG("ihsh_ars_getentry(): entryid=%s", entryid);

    /* Initialize Remedy session control block */
    control.cacheId = 0;
    strcpy(control.user, ars_user);
    strcpy(control.password, ars_passwd);
    strcpy(control.language, ""); 
    control.sessionId = 0;
    strcpy(control.server, ars_server);
    if (ARInitialization(&control, &status) >= AR_RETURN_ERROR) {
	ARS_DEBUG("ihsh_ars_getentry: ARInitialization() failed");
	if (ars_debug == TRUE) {
	    PrintARStatusList(stderr, &status);
	}
	FreeARStatusList(&status, FALSE);
        return -1;
    }
    FreeARStatusList(&status, FALSE);
    
    /* Set ARS server port */
    if (ARSetServerPort(
	    &control, ars_server, ars_port, 0, &status) >= AR_RETURN_ERROR) {
	ARS_DEBUG("ihsh_ars_getentry: ARSetServerPort() failed");
	if (ars_debug == TRUE) {
	    PrintARStatusList(stderr, &status);
	}
	FreeARStatusList(&status, FALSE);
	return -1;
    }
    FreeARStatusList(&status, FALSE);

    /* Get entry and insert it into IODEF XML tree */
    ret = ihsh_ars_getentry_iodef(&control, tree, entryid);
    if (ret < 0) {
	ARS_DEBUG("ihsh_ars_getentry(): could not get IODEF entry");
    }

    /* Terminate ARS session and clean up */
    (void) ARTermination(&control, &status);
    return ret;
}

/* Function to set entry on ARS/Remedy server */
int ihsh_ars_setentry(ih_tree_t *tree)
{
    ARControlStruct  control;
    ARStatusList     status;
    int ret;

    /* Check if IHSH ARS module is initialized */
    if (ars_initialized == FALSE) {
	ARS_DEBUG("ihsh_ars_setentry: module not initialized!");
	return -1;
    }

    ARS_DEBUG("ihsh_ars_setentry()");

    /* Initialize Remedy session control block */
    control.cacheId = 0;
    strcpy(control.user, ars_user);
    strcpy(control.password, ars_passwd);
    strcpy(control.language, ""); 
    control.sessionId = 0;
    strcpy(control.server, ars_server);
    if (ARInitialization(&control, &status) >= AR_RETURN_ERROR) {
	ARS_DEBUG("ihsh_ars_setentry: ARInitialization() failed");
	if (ars_debug == TRUE) {
	    PrintARStatusList(stderr, &status);
	}
	FreeARStatusList(&status, FALSE);
        return -1;
    }
    FreeARStatusList(&status, FALSE);
    
    /* Set ARS server port */
    if (ARSetServerPort(
	    &control, ars_server, ars_port, 0, &status) >= AR_RETURN_ERROR) {
	ARS_DEBUG("ihsh_ars_setentry: ARSetServerPort() failed");
	if (ars_debug == TRUE) {
	    PrintARStatusList(stderr, &status);
	}
	FreeARStatusList(&status, FALSE);
	return -1;
    }
    FreeARStatusList(&status, FALSE);

    /* Set IODEF XML tree contents into ARS */
    ret = ihsh_ars_setentry_iodef(&control, tree);
    if (ret < 0) {
	ARS_DEBUG("ihsh_ars_setentry(): could not set IODEF entry");
    }

    /* Terminate ARS session and clean up */
    (void) ARTermination(&control, &status);
    return ret;
}

