/*                               -*- Mode: C -*- 
 * 
 * ihsh_cfg.c -- Functions to parse IHSH configuration file
 * 
 * Copyright (c) 2003 Arne Helme. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Arne Helme.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: ihsh_cfg.c,v 1.1 2003/10/22 08:16:20 arne Exp $
 * 
 */

/* **************************************************************
 * The IHSH configuration file support is designed to be flexible
 * enough to not put any restrictions on the options being put
 * into a configration file.  Syntactically, three rules apply:
 *
 * 1. Lines beginning with an '#' are considered comments,
 * 2. empty lines are ignored,
 * 3. and the remaining lines are expected to contain an expression
 *    in the following notation <name>=<value>.
 * ************************************************************** */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ihsh_misc.h"
#include "ihsh_cfg.h"

/* Max numer of entries in an IHSH configuration file. */
#ifndef IHSH_CFG_MAXENTRIES
#define IHSH_CFG_MAXENTRIES	100
#endif

typedef struct {
    char *cfg_name;		/* Name attribue */
    char *cfg_value;		/* Associated value attribute */
} ihsh_cfg_t;

/* Local variables */

static int nentries = 0;
static ihsh_cfg_t ihsh_cfg[IHSH_CFG_MAXENTRIES];

/* Local function to insert unique entry in the config table */
static int ihsh_cfg_insertentry(char *name, char *value)
{
    int i;

    /* Check if table is full */
    if (nentries == IHSH_CFG_MAXENTRIES) {
	fprintf(stderr, "ihsh_cfg_insertentry: configuration table full!\n");
	return -1;
    }
    
    /* Check for duplicate token symbols */
    for (i = 0; i < nentries; i++) {
	if (!strcmp(name, ihsh_cfg[i].cfg_name)) {
	    fprintf(stderr, "ihsh_cfg_insertentry: duplicate symbols!\n");
	    return -1;
	}
    }
    
    /* Add new entry */
    ihsh_cfg[nentries].cfg_name = strdup(name);
    ihsh_cfg[nentries].cfg_value = strdup(value);
    nentries++;
    return ihsh_cfg[nentries - 1].cfg_name == NULL || 
	ihsh_cfg[nentries - 1].cfg_value == NULL ? -1 : 0;
}

/* Read IHSH configuration file */
int ihsh_cfg_readfile(char *fname)
{
    FILE 	*fp = NULL;
    char 	*line = NULL, *p = NULL, *q = NULL;
    int		i, ret = 0;
    
    fp = fopen(fname, "r");
    if (fp == NULL) {
	fprintf(stderr, "warning: could not open IHSH cfg file: %s\n", fname);
	return -1;
    }
    
    i = 0;
    while ((line = ihsh_getline(fp)) != NULL) {
	i++;
	ret = 0;
	/* Skip comment lines and empty lines*/
	if (line[0] == '#' || line[0] == '\n') {
	    free(line);
	    continue; /* while loop */
	}
	/* Locate first token */
	p = strtok(line, "=");
	if (p == NULL) {
	    ret = -1;
	    break;
	}
	/* Check first token */
	if (strpbrk(p, " ") != NULL) {
	    ret = -1;
	    break;
	}
	/* Locate second token */
	q = strtok(NULL, "\n\0");
	if (q == NULL) {
	    ret = -1;
	    break;
	}
	/* Check second token */
	if (strpbrk(q, " ") != NULL) {
	    ret = -1;
	    break;
	}
	/* Insert token pair into config table */
	if (ihsh_cfg_insertentry(p, q) < 0) {
	    ret = -1;
	    break;
	}
	free(line);
    }
    if (ret == -1) {
	fprintf(stderr, "%s, line %d: error: %s\n", fname, i, line);
    }
    if (fp != NULL) {
	fclose(fp);
    }
    if (line != NULL) {
	free(line);
    }
    return ret;
}

/* Return the string value attribute associated with a config name */
char *ihsh_cfg_match_str(char *name)
{
    int i;

    for (i = 0; i < nentries; i++) {
	if (!strcmp(name, ihsh_cfg[i].cfg_name)) {
	    return ihsh_cfg[i].cfg_value;
	}
    }
    return NULL;
}

/* Provide the integer value attribute associated with a config name */
int ihsh_cfg_match_int(char *name, int *num_p)
{
    long lval;
    char *str, *ep;

    str = ihsh_cfg_match_str(name);
    if (str == NULL) {
	/* Configuration parameter not found */
	return -1;
    }

    errno = 0;
    lval = strtol(str, &ep, 10);
    if (str[0] == '\0' || *ep != '\0') { /* XXX Arne, 2nd test OK ??? */
	/* Not a number */
	return -1;
    }
    if ((errno == ERANGE && (lval == LONG_MAX || lval == LONG_MIN)) ||
	(lval > INT_MAX || lval < INT_MIN)) {
	/* Number out of range */
	return -1;
    }
    *num_p = (int) lval;
    return 0;
}

/* Dump contents of configration table to stream */
void ihsh_cfg_fdump(FILE *fp)
{
    int i;

    for (i = 0; i < nentries; i++) {
	fprintf(fp, "%s=%s\n", ihsh_cfg[i].cfg_name, ihsh_cfg[i].cfg_value);
    }
}
