/*                               -*- Mode: C -*- 
 * 
 * ihsh_cmds.c -- IHSH Commands
 * 
 * Copyright (c) 2003 Arne Helme. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Arne Helme.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: ihsh_cmds.c,v 1.5 2003/10/22 13:48:46 arne Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <expat.h>

#include <ih/tree.h>
#include <ih/tree_xml.h>
#include <ih/sanitize.h>

#include "ihsh.h"
#include "ihsh_cmds.h"
#include "ihsh_misc.h"
#include "ihsh_xml.h"

/* Enumerated command types */
enum cmd {
    CMD_CD,
    CMD_DEBUG,
    CMD_ECHO,
    CMD_FSETVAL,
    CMD_GETVAL,
    CMD_HELP,
    CMD_NEW,
    CMD_NOOP,			/* No Operation, not a command */
    CMD_PWD,
    CMD_QUIT,
    CMD_READXML,
    CMD_SETVAL,
    CMD_SETATTR,
    CMD_WALK,
    CMD_WRITEXML
};

/* Command table */
struct cmdtab {
	enum cmd	command;
	const char	*name;
	unsigned int	min;
	const char	*args;
} const cmdtab[] = {
    { CMD_CD,		"cd",		1, "\t\t<Xpat-like path expr.>"},
    { CMD_DEBUG,	"debug",	1, NULL},
    { CMD_ECHO,		"echo",		1, "\t[text string]"},
    { CMD_FSETVAL,	"fsetval",	1, "\t[file name]"},
    { CMD_GETVAL,	"getval",	1, "\t\t<value>"},
    { CMD_HELP,		"?",		1, NULL},
    { CMD_HELP,		"help",		1, NULL},
    { CMD_NEW,		"new",		1, "\t\t- create a new XML tree"},
    { CMD_PWD,		"pwd",		1, "\t\t- current tree tree location"},
    { CMD_QUIT,		"quit",		1, "\t\t- exit the program"},
    { CMD_READXML,	"readxml",	1, "\t[file name]"},
    { CMD_SETVAL,	"setval",	1, "\t\t<value> [expr]"},
    { CMD_SETATTR,	"setattr",	1, "\t\t<value> [expr]"},
    { CMD_WALK,		"walk",		1, "\t\t<Xpat-like path expr.>"},
    { CMD_WRITEXML,	"writexml",	1, "\t[file name]"}
};

/* Current tree location variable */
static ih_tree_node_t *cur_node; 

/* **************************************************************
 * Run command support functions
 * ************************************************************** */

/* Help command */
static int ihsh_cmd_help(void)
{
    const struct cmdtab *c, *mc;
    const char *s;
    int i, n;

    mc = cmdtab + sizeof(cmdtab) / sizeof(cmdtab[0]);
    for (c = cmdtab; c < mc; c++) {
	for (i = c->min, s = c->name; *s != '\0'; s++, i--) {
	    n = (i > 0 ? toupper(*s) : *s);
	    putchar(n);
	}
	if (c->args != NULL) {
	    printf(" %s", c->args);
	}
	putchar('\n');
    }
    return 0;
}

/* Debug command */
static int ihsh_cmd_debug(void)
{
    debug = debug == 0 ? 1 : 0;
    return 0;
}

/* Echo command */
static int ihsh_cmd_echo(int cmd, char *arg)
{
    if (arg != NULL && arg[0] != '\0') {
	printf("%s\n", arg);
    }
    return 0;
}

/* Fsetval command */
static int ihsh_cmd_fsetval(int cmd, char *arg)
{
    FILE		*fp;
    char		*fbuf;
    size_t		nbytes, nread;
    ih_data_size_t 	len;
    ih_err_t 		result;
    
    /* Assume that the entire arg is the file name */
    if (debug) {
	printf("CMD_SETVAL: arg = '%s'\n", 
	       arg == NULL || arg[0] == '\0' ? "null" : arg);
    }
    
    fp = fopen(arg, "r");
    if (fp == NULL) {
	fprintf(stderr, "fsetval: could not open file\n");
	return -1;
    }
    (void)fseek(fp, 0L, SEEK_END);
    nbytes = (int) ftell(fp);
    if (nbytes < 0) {
	fprintf(stderr, "fsetval: could not determine file size\n");
	return -1;
    }
    fbuf = malloc(nbytes);
    if (fbuf == NULL) {
	fprintf(stderr, "fsetval: could not allocate file buffer\n");
	return -1;
    }
    (void)rewind(fp);
    nread = fread(fbuf, 1, nbytes, fp);
    if (nread <= 0) {
	fprintf(stderr, "fsetval: *FIXME*: fread must loop over file\n");
	return -1;
    }
    (void)fclose(fp);
    len = nread;
    result = ih_node_set_value(cur_node, fbuf, &len);
    free(fbuf);
    if (result != IH_E_NONE) {
	fprintf(stderr, "ih_node_set_value returned an error!\n");
	return -1;
    }
    return 0;
}

/* Getval command */
static int ihsh_cmd_getval(int cmd, char *arg)
{
    ih_err_t		result;
    ih_asciz_t		value;
    ih_data_size_t	len;
    
    result = ih_node_get_value(cur_node, &value, &len);
    if (result != IH_E_NONE) {
	if (ih_node_isset(cur_node) == IH_E_NOT_SET) {
	    printf("Node is not set\n");
	} else {
	    fprintf(stderr, "ih_node_get_value returned an error\n");
	}
	return -1;
    }
    printf("%s\n", value);
    return 0;
}

/* New command */
static int ihsh_cmd_new(ih_tree_t *tree, ualloc_t *ua)
{
    ih_err_t 	result;

    result = ih_tree_destroy(tree);
    if (result != IH_E_NONE) {
	fprintf(stderr, "Error: tree_destroy: result = %d\n", result);
	return -1;
    }
    tree = ih_tree_create(ua, IH_TREE_T_IODEF);
    if (tree == NULL) {
	fprintf(stderr, "Error: could not create tree\n");
	return -1;
    }
    return 0;
}

/* Pwd command */
static int ihsh_cmd_pwd(void)
{
    fprintf(stderr, "pwd not yet implemented!\n");
    /* printf("%s\n", ih_node_path(cur_node, 0)); */
    return 0;
}

/* Readxml command */
static int ihsh_cmd_readxml(ih_tree_t **tree_p, int cmd, char *arg, 
			    ualloc_t *ua)
{
    FILE	*fp;
    int		ret;

    ret = -1;
    fp = fopen(arg, "r");
    if (fp != NULL) {
	ret = ihsh_xml_fread(tree_p, fp, ua);
	fclose(fp);
    }
    if (ret < 0) {
	fprintf(stderr, "Error: could not read XML file\n");
    }
    return ret;
}

/* Setval command */
static int ihsh_cmd_setval(int cmd, char *arg, FILE *fp)
{
    char		*buf;
    ih_err_t		result;
    ih_data_size_t	len;

    /* Assume that the entire arg is the content to be set */
    if (debug) {
	printf("CMD_SETVAL: arg = '%s'\n", 
	       arg == NULL || arg[0] == '\0' ? "null" : arg);
    }
    if (arg != NULL && arg[0] != '\0') {
	len = strlen(arg);
	result = ih_node_set_value(cur_node, arg, &len);
	if (result != IH_E_NONE) {
	    fprintf(stderr, "ih_node_set_value returned an error!\n");
	    return -1;
	}
    } else {
	/* no argument to SETVAL given, assume multi-line input */
	buf = ihsh_getlines(fp);
	if (buf == NULL) {
	    fprintf(stderr, "could not read multi-line input!\n");
	    return -1;
	}
	len = strlen(buf);
	result = ih_node_set_value(cur_node, buf, &len);
	free(buf);
	if (result != IH_E_NONE) {
	    fprintf(stderr, "ih_node_set_value returned an error!\n");
	    return -1;
	}
    }
    return 0;
}

/* Setattr command */
static int ihsh_cmd_setattr(int cmd, char *arg)
{
    ih_err_t		result;
    ih_data_size_t	len;
    char 		*attrname, *attrval;
    
    if (arg == NULL || arg[0] == '\0') {
	return -1;
    }
    
    /* In case of a setattr call, arg is expected to contain the 
       attribute name plus the attributed value(s) */
    attrname = strdup(arg);
    if (attrname == NULL) {
	/* malloc error */
	return -1;
    }
	    
    attrval = strchr(attrname, ' ');
    if (attrval == NULL) {
	/* no attribute value found */
	free(attrname);
	return -1;
    }
    *attrval++ = '\0';
    len = strlen(attrval);

    result = ih_node_attr_set_by_name(cur_node, attrname, attrval, &len);
    free(attrname);
    if (result != IH_E_NONE) {
	fprintf(stderr, "ih_node_attr_set_by_name returned an error!\n");
	return -1;
    }
    return 0;
}

/* Walk command */
static int ihsh_cmd_walk(ih_tree_t *tree, int cmd, char *arg)
{
    ih_err_t 	result;

    /*
     * TODO: walking relative paths still needs to be fixed.
     */
    if (!strcmp(arg, "..")) {
	ih_tree_node_t *parent_node;

	result = ih_node_get_parent(cur_node, &parent_node);
	if (result != IH_E_NONE) {
	    fprintf(stderr, 
		    "Error: ih_node_get_parent: %s\n",
		    ih_err_code2str(result));
	    return -1;
	}
	cur_node = parent_node;
    } else {
	result = ih_tree_walk_by_path2(tree->t_root, arg, &cur_node);
	if (result != IH_E_NONE) {
	    fprintf(stderr, 
		    "Error: ihsh_walk_by_path: %s\n",
		    ih_err_code2str(result));
	    return -1;
	}
    }
    if (debug) {
	printf("Current node type = %s\n", 
	       ih_node_type2str(cur_node->n_type));
    }
    return 0;
}

/* Writexml command */
static int ihsh_cmd_writexml(ih_tree_t *tree, int cmd, char *arg)
{
    FILE	*fp;
    int		ret;

    ret = -1;
    if (arg != NULL && arg[0] != '\0') {
	fp = fopen(arg, "w");
	if (fp == NULL) {
	    fprintf(stderr, "CMD_WRITEXML: could not open '%s'\n", arg);
	    return -1;
	}
    } else {
	fp = stdout;
    }
    ret = ihsh_xml_fwrite(tree, fp);
    if (fp != stdout) {
	fclose(fp);
    }
    if (ret < 0) {
	fprintf(stderr, "Error: could not write XML file\n");
    }
    return ret;
}

/* Function to run IHSH command */
static int ihsh_cmd(ih_tree_t **tree_p, ualloc_t *ua, 
		    int cmd, char *arg, FILE *fp)
{
    int ret;

    if (cmd == CMD_QUIT) {
	exit(EXIT_SUCCESS);
	/* NOTREACHED */
    }

    switch (cmd) {
    case CMD_DEBUG:
	ret = ihsh_cmd_debug();
	break;

    case CMD_ECHO:
	ret = ihsh_cmd_echo(cmd, arg);
	break;

    case CMD_FSETVAL:
	ret = ihsh_cmd_fsetval(cmd, arg);
	break;

    case CMD_GETVAL:
	ret = ihsh_cmd_getval(cmd, arg);
	break;

    case CMD_NEW:
	ret = ihsh_cmd_new(*tree_p, ua);
	break;

    case CMD_NOOP:
	/* Noop, do nothing */
	ret = 0;
	break;

    case CMD_PWD:
	ret = ihsh_cmd_pwd();
	break;

    case CMD_READXML:
	ret = ihsh_cmd_readxml(tree_p, cmd, arg, ua);
	break;

    case CMD_SETVAL:
	ret = ihsh_cmd_setval(cmd, arg, fp);
	break;

    case CMD_SETATTR:
	ret = ihsh_cmd_setattr(cmd, arg);
	break;

    case CMD_CD:
    case CMD_WALK:
	ret = ihsh_cmd_walk(*tree_p, cmd, arg);
	break;

    case CMD_WRITEXML:
	ret = ihsh_cmd_writexml(*tree_p, cmd, arg);
	break;
	
    case CMD_HELP:
	ret = ihsh_cmd_help();
	break;
    default:
	break;
    }
    return ret;
}

/* Function to parse IHSH command */
static char *ihsh_parse(char *buf, int *cmd)
{
    const struct cmdtab *c, *mc;
    char *p, *q;
    int len;

    for (p = buf; isspace((int)*p); p++) {
	continue;
    }
    for (buf = p; *p != '\0' && !isspace((int)*p); p++) {
	continue;
    }
    if ((len = p - buf) == 0) {
	return NULL;
    }
    if (*p != '\0') {		/* It must be a spacing character! */
	*p++ = 0;
	for (q = p; *q != '\0' && *q != '\n' && *q != '\r'; q++)
	    continue;
	*q = 0;
    }

    *cmd = -1;

    mc = cmdtab + sizeof(cmdtab) / sizeof(cmdtab[0]);
    for (c = cmdtab; c < mc; c++) {
	/* Is it an exact match? */
	if (strcasecmp(buf, c->name) == 0) {
	    *cmd = c->command;
	    break;
	}
	/* Try short hand forms then... */
	if (len >= c->min && strncasecmp(buf, c->name, len) == 0) {
	    if (*cmd != -1 && *cmd != c->command) {
		fprintf(stderr, "ambiguous command\n");
		return (0);
	    }
	    *cmd = c->command;
	}
    }

    if (*cmd == -1) {
	return NULL;
    }

    while (isspace((int)*p)) {
	p++;
    }
    return p;
}

/* Function to process IHSH file command stream */
void ihsh_cmd_fproc(FILE *fp, ih_tree_t **tree_p, ualloc_t *ua)
{
    int 	cmd;
    char 	*line;
    char 	*arg;

    while (!feof(fp)) {
	line = NULL;
	do {
	    if ((line = ihsh_getline(fp)) != NULL) {
		if (line[0] != '#') {
		    arg = ihsh_parse(line, &cmd);	
		    if (cmd < 0) {
			fprintf(stderr, "invalid input: %s, cmd: %i\n",
				line, cmd);
			exit(EXIT_FAILURE);
		    }
		} else {
		    cmd = CMD_NOOP;
		    arg = NULL;	
		    
		}
	    }
	} while (arg == NULL && !feof(fp));
	if (arg != NULL) {
	    if (ihsh_cmd(tree_p, ua, cmd, arg, fp) < 0) {
		fprintf(stderr, "ihsh command error!\n");
	    }
	} 
	if (line != NULL) {
	    free(line); 
	}
    }
}
